/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.datamodel;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.BitSet;
import java.util.List;

/**
 * Holds a list of search result matches, where each match is a contiguous
 * stretch of a single sequence.
 * 
 * @author gmcarstairs amwaterhouse
 *
 */
public class SearchResults implements SearchResultsI
{

  private List<SearchResultMatchI> matches = new ArrayList<SearchResultMatchI>();

  /**
   * One match consists of a sequence reference, start and end positions.
   * Discontiguous ranges in a sequence require two or more Match objects.
   */
  public class Match implements SearchResultMatchI
  {
    SequenceI sequence;

    /**
     * Start position of match in sequence (base 1)
     */
    int start;

    /**
     * End position (inclusive) (base 1)
     */
    int end;

    /**
     * Constructor
     * 
     * @param seq
     *          a sequence
     * @param start
     *          start position of matched range (base 1)
     * @param end
     *          end of matched range (inclusive, base 1)
     */
    public Match(SequenceI seq, int start, int end)
    {
      sequence = seq;

      /*
       * always hold in forwards order, even if given in reverse order
       * (such as from a mapping to a reverse strand); this avoids
       * trouble for routines that highlight search results etc
       */
      if (start <= end)
      {
        this.start = start;
        this.end = end;
      }
      else
      {
        this.start = end;
        this.end = start;
      }
    }

    /* (non-Javadoc)
     * @see jalview.datamodel.SearchResultMatchI#getSequence()
     */
    @Override
    public SequenceI getSequence()
    {
      return sequence;
    }

    /* (non-Javadoc)
     * @see jalview.datamodel.SearchResultMatchI#getStart()
     */
    @Override
    public int getStart()
    {
      return start;
    }

    /* (non-Javadoc)
     * @see jalview.datamodel.SearchResultMatchI#getEnd()
     */
    @Override
    public int getEnd()
    {
      return end;
    }

    /**
     * Returns the string of characters in the matched region, prefixed by the
     * start position, e.g. "12CGT" or "208K"
     */
    @Override
    public String toString()
    {
      final int from = Math.max(start - 1, 0);
      String startPosition = String.valueOf(from);
      return startPosition + getCharacters();
    }

    /* (non-Javadoc)
     * @see jalview.datamodel.SearchResultMatchI#getCharacters()
     */
    @Override
    public String getCharacters()
    {
      char[] chars = sequence.getSequence();
      // convert start/end to base 0 (with bounds check)
      final int from = Math.max(start - 1, 0);
      final int to = Math.min(end, chars.length + 1);
      return String.valueOf(Arrays.copyOfRange(chars, from, to));
    }

    public void setSequence(SequenceI seq)
    {
      this.sequence = seq;
    }

    /**
     * Hashcode is the hashcode of the matched sequence plus a hash of start and
     * end positions. Match objects that pass the test for equals are guaranteed
     * to have the same hashcode.
     */
    @Override
    public int hashCode()
    {
      int hash = sequence == null ? 0 : sequence.hashCode();
      hash += 31 * start;
      hash += 67 * end;
      return hash;
    }

    /**
     * Two Match objects are equal if they are for the same sequence, start and
     * end positions
     */
    @Override
    public boolean equals(Object obj)
    {
      if (obj == null || !(obj instanceof SearchResultMatchI))
      {
        return false;
      }
      SearchResultMatchI m = (SearchResultMatchI) obj;
      return (sequence == m.getSequence() && start == m.getStart() && end == m
              .getEnd());
    }
  }

  /* (non-Javadoc)
   * @see jalview.datamodel.SearchResultsI#addResult(jalview.datamodel.SequenceI, int, int)
   */
  @Override
  public SearchResultMatchI addResult(SequenceI seq, int start, int end)
  {
    Match m = new Match(seq, start, end);
    matches.add(m);
    return m;
  }

  /* (non-Javadoc)
   * @see jalview.datamodel.SearchResultsI#involvesSequence(jalview.datamodel.SequenceI)
   */
  @Override
  public boolean involvesSequence(SequenceI sequence)
  {
    SequenceI ds = sequence.getDatasetSequence();
    Match m;
    for (SearchResultMatchI _m : matches)
    {
      m = (Match) _m;
      if (m.sequence != null
              && (m.sequence == sequence || m.sequence == ds))
      {
        return true;
      }
    }
    return false;
  }

  /* (non-Javadoc)
   * @see jalview.datamodel.SearchResultsI#getResults(jalview.datamodel.SequenceI, int, int)
   */
  @Override
  public int[] getResults(SequenceI sequence, int start, int end)
  {
    if (matches.isEmpty())
    {
      return null;
    }

    int[] result = null;
    int[] tmp = null;
    int resultLength, matchStart = 0, matchEnd = 0;
    boolean mfound;
    Match m;
    for (SearchResultMatchI _m : matches)
    {
      m = (Match) _m;

      mfound = false;
      if (m.sequence == sequence)
      {
        mfound = true;
        // locate aligned position
        matchStart = sequence.findIndex(m.start) - 1;
        matchEnd = sequence.findIndex(m.end) - 1;
      }
      else if (m.sequence == sequence.getDatasetSequence())
      {
        mfound = true;
        // locate region in local context
        matchStart = sequence.findIndex(m.start) - 1;
        matchEnd = sequence.findIndex(m.end) - 1;
      }
      if (mfound)
      {
        if (matchStart <= end && matchEnd >= start)
        {
          if (matchStart < start)
          {
            matchStart = start;
          }

          if (matchEnd > end)
          {
            matchEnd = end;
          }

          if (result == null)
          {
            result = new int[] { matchStart, matchEnd };
          }
          else
          {
            resultLength = result.length;
            tmp = new int[resultLength + 2];
            System.arraycopy(result, 0, tmp, 0, resultLength);
            result = tmp;
            result[resultLength] = matchStart;
            result[resultLength + 1] = matchEnd;
          }
        }
        else
        {
          // debug
          // System.err.println("Outwith bounds!" + matchStart+">"+end +"  or "
          // + matchEnd+"<"+start);
        }
      }
    }
    return result;
  }

  @Override
  public int markColumns(SequenceCollectionI sqcol, BitSet bs)
  {
    int count = 0;
    BitSet mask = new BitSet();
    for (SequenceI s : sqcol.getSequences())
    {
      int[] cols = getResults(s, sqcol.getStartRes(), sqcol.getEndRes());
      if (cols != null)
      {
        for (int pair = 0; pair < cols.length; pair += 2)
        {
          mask.set(cols[pair], cols[pair + 1] + 1);
        }
      }
    }
    // compute columns that were newly selected
    BitSet original = (BitSet) bs.clone();
    original.and(mask);
    count = mask.cardinality() - original.cardinality();
    // and mark ranges not already marked
    bs.or(mask);
    return count;
  }

  /* (non-Javadoc)
   * @see jalview.datamodel.SearchResultsI#getSize()
   */
  @Override
  public int getSize()
  {
    return matches.size();
  }

  /* (non-Javadoc)
   * @see jalview.datamodel.SearchResultsI#getResultSequence(int)
   */
  @Override
  public SequenceI getResultSequence(int index)
  {
    return matches.get(index).getSequence();
  }

  /* (non-Javadoc)
   * @see jalview.datamodel.SearchResultsI#getResultStart(int)
   */
  @Override
  public int getResultStart(int i)
  {
    return matches.get(i).getStart();
  }

  /* (non-Javadoc)
   * @see jalview.datamodel.SearchResultsI#getResultEnd(int)
   */
  @Override
  public int getResultEnd(int i)
  {
    return matches.get(i).getEnd();
  }

  /* (non-Javadoc)
   * @see jalview.datamodel.SearchResultsI#isEmpty()
   */
  @Override
  public boolean isEmpty()
  {
    return matches.isEmpty();
  }

  /* (non-Javadoc)
   * @see jalview.datamodel.SearchResultsI#getResults()
   */
  @Override
  public List<SearchResultMatchI> getResults()
  {
    return matches;
  }

  /**
   * Return the results as a string of characters (bases) prefixed by start
   * position(s). Meant for use when the context ensures that all matches are to
   * regions of the same sequence (otherwise the result is meaningless).
   * 
   * @return
   */
  @Override
  public String toString()
  {
    StringBuilder result = new StringBuilder(256);
    for (SearchResultMatchI m : matches)
    {
      result.append(m.toString());
    }
    return result.toString();
  }

  /**
   * Return the results as a string of characters (bases). Meant for use when
   * the context ensures that all matches are to regions of the same sequence
   * (otherwise the result is meaningless).
   * 
   * @return
   */
  public String getCharacters()
  {
    StringBuilder result = new StringBuilder(256);
    for (SearchResultMatchI m : matches)
    {
      result.append(m.getCharacters());
    }
    return result.toString();
  }

  /**
   * Hashcode is has derived from the list of matches. This ensures that when
   * two SearchResults objects satisfy the test for equals(), then they have the
   * same hashcode.
   */
  @Override
  public int hashCode()
  {
    return matches.hashCode();
  }

  /**
   * Two SearchResults are considered equal if they contain the same matches in
   * the same order.
   */
  @Override
  public boolean equals(Object obj)
  {
    if (obj == null || !(obj instanceof SearchResultsI))
    {
      return false;
    }
    SearchResultsI sr = (SearchResultsI) obj;
    return matches.equals(sr.getResults());
  }
}
